@@ -1,22 +1,22 @@
 from functools import update_wrapper
+from weakref import WeakSet
 
 from django.apps import apps
-from django.conf import settings
 from django.contrib.admin import ModelAdmin, actions
 from django.contrib.auth import REDIRECT_FIELD_NAME
-from django.core.exceptions import ImproperlyConfigured, PermissionDenied
-from django.core.urlresolvers import NoReverseMatch, reverse
+from django.core.exceptions import ImproperlyConfigured
 from django.db.models.base import ModelBase
 from django.http import Http404, HttpResponseRedirect
-from django.template.engine import Engine
 from django.template.response import TemplateResponse
+from django.urls import NoReverseMatch, reverse
 from django.utils import six
 from django.utils.text import capfirst
 from django.utils.translation import ugettext as _, ugettext_lazy
 from django.views.decorators.cache import never_cache
 from django.views.decorators.csrf import csrf_protect
+from django.views.i18n import JavaScriptCatalog
 
-system_check_errors = []
+all_sites = WeakSet()
 
 
 class AlreadyRegistered(Exception):
@@ -48,6 +48,8 @@ class AdminSite(object):
     # URL for the "View site" link at the top of each admin page.
     site_url = '/'
 
+    _empty_value_display = '-'
+
     login_form = None
     index_template = None
     app_index_template = None
@@ -61,6 +63,23 @@ class AdminSite(object):
         self.name = name
         self._actions = {'delete_selected': actions.delete_selected}
         self._global_actions = self._actions.copy()
+        all_sites.add(self)
+
+    def check(self, app_configs):
+        """
+        Run the system checks on all ModelAdmins, except if they aren't
+        customized at all.
+        """
+        if app_configs is None:
+            app_configs = apps.get_app_configs()
+        app_configs = set(app_configs)  # Speed up lookups below
+
+        errors = []
+        modeladmins = (o for o in self._registry.values() if o.__class__ is not ModelAdmin)
+        for modeladmin in modeladmins:
+            if modeladmin.model._meta.app_config in app_configs:
+                errors.extend(modeladmin.check())
+        return errors
 
     def register(self, model_or_iterable, admin_class=None, **options):
         """
@@ -83,8 +102,9 @@ class AdminSite(object):
             model_or_iterable = [model_or_iterable]
         for model in model_or_iterable:
             if model._meta.abstract:
-                raise ImproperlyConfigured('The model %s is abstract, so it '
-                      'cannot be registered with admin.' % model.__name__)
+                raise ImproperlyConfigured(
+                    'The model %s is abstract, so it cannot be registered with admin.' % model.__name__
+                )
 
             if model in self._registry:
                 raise AlreadyRegistered('The model %s is already registered' % model.__name__)
@@ -101,9 +121,6 @@ class AdminSite(object):
                     options['__module__'] = __name__
                     admin_class = type("%sAdmin" % model.__name__, (admin_class,), options)
 
-                if admin_class is not ModelAdmin and settings.DEBUG:
-                    system_check_errors.extend(admin_class.check(model))
-
                 # Instantiate the admin class to save in the registry
                 self._registry[model] = admin_class(model, self)
 
@@ -154,6 +171,14 @@ class AdminSite(object):
         """
         return six.iteritems(self._actions)
 
+    @property
+    def empty_value_display(self):
+        return self._empty_value_display
+
+    @empty_value_display.setter
+    def empty_value_display(self, empty_value_display):
+        self._empty_value_display = empty_value_display
+
     def has_permission(self, request):
         """
         Returns True if the given HttpRequest has permission to view
@@ -161,40 +186,6 @@ class AdminSite(object):
         """
         return request.user.is_active and request.user.is_staff
 
-    def check_dependencies(self):
-        """
-        Check that all things needed to run the admin have been correctly installed.
-
-        The default implementation checks that admin and contenttypes apps are
-        installed, as well as the auth context processor.
-        """
-        if not apps.is_installed('django.contrib.admin'):
-            raise ImproperlyConfigured(
-                "Put 'django.contrib.admin' in your INSTALLED_APPS "
-                "setting in order to use the admin application.")
-        if not apps.is_installed('django.contrib.contenttypes'):
-            raise ImproperlyConfigured(
-                "Put 'django.contrib.contenttypes' in your INSTALLED_APPS "
-                "setting in order to use the admin application.")
-        try:
-            default_template_engine = Engine.get_default()
-        except Exception:
-            # Skip this non-critical check:
-            # 1. if the user has a non-trivial TEMPLATES setting and Django
-            #    can't find a default template engine
-            # 2. if anything goes wrong while loading template engines, in
-            #    order to avoid raising an exception from a confusing location
-            # Catching ImproperlyConfigured suffices for 1. but 2. requires
-            # catching all exceptions.
-            pass
-        else:
-            if ('django.contrib.auth.context_processors.auth'
-                    not in default_template_engine.context_processors):
-                raise ImproperlyConfigured(
-                    "Enable 'django.contrib.auth.context_processors.auth' "
-                    "in your TEMPLATES setting in order to use the admin "
-                    "application.")
-
     def admin_view(self, view, cacheable=False):
         """
         Decorator to create an admin view attached to this ``AdminSite``. This
@@ -246,12 +237,10 @@ class AdminSite(object):
         # and django.contrib.contenttypes.views imports ContentType.
         from django.contrib.contenttypes import views as contenttype_views
 
-        if settings.DEBUG:
-            self.check_dependencies()
-
         def wrap(view, cacheable=False):
             def wrapper(*args, **kwargs):
                 return self.admin_view(view, cacheable)(*args, **kwargs)
+            wrapper.admin_site = self
             return update_wrapper(wrapper, view)
 
         # Admin-site-wide views.
@@ -270,7 +259,7 @@ class AdminSite(object):
         # Add in each model's views, and create a list of valid URLS for the
         # app_index
         valid_app_labels = []
-        for model, model_admin in six.iteritems(self._registry):
+        for model, model_admin in self._registry.items():
             urlpatterns += [
                 url(r'^%s/%s/' % (model._meta.app_label, model._meta.model_name), include(model_admin.urls)),
             ]
@@ -294,12 +283,18 @@ class AdminSite(object):
         """
         Returns a dictionary of variables to put in the template context for
         *every* page in the admin site.
+
+        For sites running on a subpath, use the SCRIPT_NAME value if site_url
+        hasn't been customized.
         """
+        script_name = request.META['SCRIPT_NAME']
+        site_url = script_name if self.site_url == '/' and script_name else self.site_url
         return {
             'site_title': self.site_title,
             'site_header': self.site_header,
-            'site_url': self.site_url,
+            'site_url': site_url,
             'has_permission': self.has_permission(request),
+            'available_apps': self.get_app_list(request),
         }
 
     def password_change(self, request, extra_context=None):
@@ -307,43 +302,39 @@ class AdminSite(object):
         Handles the "change password" task -- both form display and validation.
         """
         from django.contrib.admin.forms import AdminPasswordChangeForm
-        from django.contrib.auth.views import password_change
+        from django.contrib.auth.views import PasswordChangeView
         url = reverse('admin:password_change_done', current_app=self.name)
         defaults = {
-            'current_app': self.name,
-            'password_change_form': AdminPasswordChangeForm,
-            'post_change_redirect': url,
+            'form_class': AdminPasswordChangeForm,
+            'success_url': url,
             'extra_context': dict(self.each_context(request), **(extra_context or {})),
         }
         if self.password_change_template is not None:
             defaults['template_name'] = self.password_change_template
-        return password_change(request, **defaults)
+        request.current_app = self.name
+        return PasswordChangeView.as_view(**defaults)(request)
 
     def password_change_done(self, request, extra_context=None):
         """
         Displays the "success" page after a password change.
         """
-        from django.contrib.auth.views import password_change_done
+        from django.contrib.auth.views import PasswordChangeDoneView
         defaults = {
-            'current_app': self.name,
             'extra_context': dict(self.each_context(request), **(extra_context or {})),
         }
         if self.password_change_done_template is not None:
             defaults['template_name'] = self.password_change_done_template
-        return password_change_done(request, **defaults)
+        request.current_app = self.name
+        return PasswordChangeDoneView.as_view(**defaults)(request)
 
-    def i18n_javascript(self, request):
+    def i18n_javascript(self, request, extra_context=None):
         """
         Displays the i18n JavaScript that the Django admin requires.
 
-        This takes into account the USE_I18N setting. If it's set to False, the
-        generated JavaScript will be leaner and faster.
+        `extra_context` is unused but present for consistency with the other
+        admin views.
         """
-        if settings.USE_I18N:
-            from django.views.i18n import javascript_catalog
-        else:
-            from django.views.i18n import null_javascript_catalog as javascript_catalog
-        return javascript_catalog(request, packages=['django.conf', 'django.contrib.admin'])
+        return JavaScriptCatalog.as_view(packages=['django.contrib.admin'])(request)
 
     @never_cache
     def logout(self, request, extra_context=None):
@@ -352,9 +343,8 @@ class AdminSite(object):
 
         This should *not* assume the user is already logged in.
         """
-        from django.contrib.auth.views import logout
+        from django.contrib.auth.views import LogoutView
         defaults = {
-            'current_app': self.name,
             'extra_context': dict(
                 self.each_context(request),
                 # Since the user isn't logged out at this point, the value of
@@ -365,7 +355,8 @@ class AdminSite(object):
         }
         if self.logout_template is not None:
             defaults['template_name'] = self.logout_template
-        return logout(request, **defaults)
+        request.current_app = self.name
+        return LogoutView.as_view(**defaults)(request)
 
     @never_cache
     def login(self, request, extra_context=None):
@@ -377,27 +368,110 @@ class AdminSite(object):
             index_path = reverse('admin:index', current_app=self.name)
             return HttpResponseRedirect(index_path)
 
-        from django.contrib.auth.views import login
+        from django.contrib.auth.views import LoginView
         # Since this module gets imported in the application's root package,
         # it cannot import models from other applications at the module level,
         # and django.contrib.admin.forms eventually imports User.
         from django.contrib.admin.forms import AdminAuthenticationForm
-        context = dict(self.each_context(request),
+        context = dict(
+            self.each_context(request),
             title=_('Log in'),
             app_path=request.get_full_path(),
+            username=request.user.get_username(),
         )
         if (REDIRECT_FIELD_NAME not in request.GET and
                 REDIRECT_FIELD_NAME not in request.POST):
-            context[REDIRECT_FIELD_NAME] = request.get_full_path()
+            context[REDIRECT_FIELD_NAME] = reverse('admin:index', current_app=self.name)
         context.update(extra_context or {})
 
         defaults = {
             'extra_context': context,
-            'current_app': self.name,
             'authentication_form': self.login_form or AdminAuthenticationForm,
             'template_name': self.login_template or 'admin/login.html',
         }
-        return login(request, **defaults)
+        request.current_app = self.name
+        return LoginView.as_view(**defaults)(request)
+
+    def _build_app_dict(self, request, label=None):
+        """
+        Builds the app dictionary. Takes an optional label parameters to filter
+        models of a specific app.
+        """
+        app_dict = {}
+
+        if label:
+            models = {
+                m: m_a for m, m_a in self._registry.items()
+                if m._meta.app_label == label
+            }
+        else:
+            models = self._registry
+
+        for model, model_admin in models.items():
+            app_label = model._meta.app_label
+
+            has_module_perms = model_admin.has_module_permission(request)
+            if not has_module_perms:
+                continue
+
+            perms = model_admin.get_model_perms(request)
+
+            # Check whether user has any perm for this module.
+            # If so, add the module to the model_list.
+            if True not in perms.values():
+                continue
+
+            info = (app_label, model._meta.model_name)
+            model_dict = {
+                'name': capfirst(model._meta.verbose_name_plural),
+                'object_name': model._meta.object_name,
+                'perms': perms,
+            }
+            if perms.get('change'):
+                try:
+                    model_dict['admin_url'] = reverse('admin:%s_%s_changelist' % info, current_app=self.name)
+                except NoReverseMatch:
+                    pass
+            if perms.get('add'):
+                try:
+                    model_dict['add_url'] = reverse('admin:%s_%s_add' % info, current_app=self.name)
+                except NoReverseMatch:
+                    pass
+
+            if app_label in app_dict:
+                app_dict[app_label]['models'].append(model_dict)
+            else:
+                app_dict[app_label] = {
+                    'name': apps.get_app_config(app_label).verbose_name,
+                    'app_label': app_label,
+                    'app_url': reverse(
+                        'admin:app_list',
+                        kwargs={'app_label': app_label},
+                        current_app=self.name,
+                    ),
+                    'has_module_perms': has_module_perms,
+                    'models': [model_dict],
+                }
+
+        if label:
+            return app_dict.get(label)
+        return app_dict
+
+    def get_app_list(self, request):
+        """
+        Returns a sorted list of all the installed apps that have been
+        registered in this site.
+        """
+        app_dict = self._build_app_dict(request)
+
+        # Sort the apps alphabetically.
+        app_list = sorted(app_dict.values(), key=lambda x: x['name'].lower())
+
+        # Sort the models alphabetically within each app.
+        for app in app_list:
+            app['models'].sort(key=lambda x: x['name'])
+
+        return app_list
 
     @never_cache
     def index(self, request, extra_context=None):
@@ -405,55 +479,7 @@ class AdminSite(object):
         Displays the main admin index page, which lists all of the installed
         apps that have been registered in this site.
         """
-        app_dict = {}
-        for model, model_admin in self._registry.items():
-            app_label = model._meta.app_label
-            has_module_perms = model_admin.has_module_permission(request)
-
-            if has_module_perms:
-                perms = model_admin.get_model_perms(request)
-
-                # Check whether user has any perm for this module.
-                # If so, add the module to the model_list.
-                if True in perms.values():
-                    info = (app_label, model._meta.model_name)
-                    model_dict = {
-                        'name': capfirst(model._meta.verbose_name_plural),
-                        'object_name': model._meta.object_name,
-                        'perms': perms,
-                    }
-                    if perms.get('change', False):
-                        try:
-                            model_dict['admin_url'] = reverse('admin:%s_%s_changelist' % info, current_app=self.name)
-                        except NoReverseMatch:
-                            pass
-                    if perms.get('add', False):
-                        try:
-                            model_dict['add_url'] = reverse('admin:%s_%s_add' % info, current_app=self.name)
-                        except NoReverseMatch:
-                            pass
-                    if app_label in app_dict:
-                        app_dict[app_label]['models'].append(model_dict)
-                    else:
-                        app_dict[app_label] = {
-                            'name': apps.get_app_config(app_label).verbose_name,
-                            'app_label': app_label,
-                            'app_url': reverse(
-                                'admin:app_list',
-                                kwargs={'app_label': app_label},
-                                current_app=self.name,
-                            ),
-                            'has_module_perms': has_module_perms,
-                            'models': [model_dict],
-                        }
-
-        # Sort the apps alphabetically.
-        app_list = list(six.itervalues(app_dict))
-        app_list.sort(key=lambda x: x['name'].lower())
-
-        # Sort the models alphabetically within each app.
-        for app in app_list:
-            app['models'].sort(key=lambda x: x['name'])
+        app_list = self.get_app_list(request)
 
         context = dict(
             self.each_context(request),
@@ -464,57 +490,17 @@ class AdminSite(object):
 
         request.current_app = self.name
 
-        return TemplateResponse(request, self.index_template or
-                                'admin/index.html', context)
+        return TemplateResponse(request, self.index_template or 'admin/index.html', context)
 
     def app_index(self, request, app_label, extra_context=None):
-        app_name = apps.get_app_config(app_label).verbose_name
-        app_dict = {}
-        for model, model_admin in self._registry.items():
-            if app_label == model._meta.app_label:
-                has_module_perms = model_admin.has_module_permission(request)
-                if not has_module_perms:
-                    raise PermissionDenied
-
-                perms = model_admin.get_model_perms(request)
-
-                # Check whether user has any perm for this module.
-                # If so, add the module to the model_list.
-                if True in perms.values():
-                    info = (app_label, model._meta.model_name)
-                    model_dict = {
-                        'name': capfirst(model._meta.verbose_name_plural),
-                        'object_name': model._meta.object_name,
-                        'perms': perms,
-                    }
-                    if perms.get('change'):
-                        try:
-                            model_dict['admin_url'] = reverse('admin:%s_%s_changelist' % info, current_app=self.name)
-                        except NoReverseMatch:
-                            pass
-                    if perms.get('add'):
-                        try:
-                            model_dict['add_url'] = reverse('admin:%s_%s_add' % info, current_app=self.name)
-                        except NoReverseMatch:
-                            pass
-                    if app_dict:
-                        app_dict['models'].append(model_dict),
-                    else:
-                        # First time around, now that we know there's
-                        # something to display, add in the necessary meta
-                        # information.
-                        app_dict = {
-                            'name': app_name,
-                            'app_label': app_label,
-                            'app_url': '',
-                            'has_module_perms': has_module_perms,
-                            'models': [model_dict],
-                        }
+        app_dict = self._build_app_dict(request, app_label)
         if not app_dict:
             raise Http404('The requested admin page does not exist.')
         # Sort the models alphabetically within each app.
         app_dict['models'].sort(key=lambda x: x['name'])
-        context = dict(self.each_context(request),
+        app_name = apps.get_app_config(app_label).verbose_name
+        context = dict(
+            self.each_context(request),
             title=_('%(app)s administration') % {'app': app_name},
             app_list=[app_dict],
             app_label=app_label,
@@ -528,6 +514,7 @@ class AdminSite(object):
             'admin/app_index.html'
         ], context)
 
+
 # This global object represents the default admin site, for the common case.
 # You can instantiate AdminSite in your own code to create a custom admin site.
 site = AdminSite()
